
from . import tools
from . import helpers


__all__ = ('tables', 'sculpt')


tables = {}


def table(new, name, *args):

    """
    Create the structure and its respective table.
    """

    cls = new.__getattr__(name)

    tables[cls] = tools.Table(*args)

    return cls


def sculpt(new, *assets):

    """
    Create structures and link them to handling instructions.

    .. code-block::

        sculpt(
            New(),
            (
                'Item',
                {
                    Key(1, 'name'): field(str),
                    Key(1, 'broken'): field(bool)
                }
            ),
            (
                'User',
                {
                    Key(0, 'id'): field(int),
                    Key(1, 'name'): field(str),
                    Key(2, 'invetory'): field(Array, Item)
                }
            )
        )
    """

    for asset in assets:

        yield table(new, *asset)



import functools


__all__ = ('resolve',)


def imitate(value, name = None, module = None, doc = None):

    if name:

        value.__name__ = value.__qualname__ = name

    if module:

        value.__module__ = module

    if doc:

        value.__doc__ = doc


def simulate(cls, /, name = None, module = None, doc = None):

    class meta(type):

        __subclasscheck__ = cls.__subclasscheck__

        __instancecheck__ = cls.__instancecheck__

        __bases__ = (cls,)

    class value(metaclass = meta):

        """.."""

        __slots__ = ()

        def __new__(cls, *args, **kwargs):

            return cls.__bases__[0](*args, **kwargs)

    imitate(value, name = name or cls.__name__, module = module, doc = doc)

    return value


def resolve(value):

    if isinstance(value, functools.partial):

        return resolve(value.func)

    return value



import math
import functools
import collections

from . import cache
from . import helpers
from . import utils


__all__ = ('Collect', 'Array', 'Mapping', 'Structure')


class Collect:

    """
    Manager for building and limiting collections.

    :param func factory:
        Used for creating new values with incoming data.
    :param int limit:
        Forget old values when adding new ones that exceed this.
    :param bool clean:
        Whether to clear all values before bulk-updating with new ones.
    """

    def __init__(self, factory, *, limit = math.inf, clean = True):

        self._factory = factory

        self._limit = limit

        self._clean = clean

    @property
    def factory(self):

        return self._factory

    @property
    def limit(self):

        return self._limit

    @limit.setter
    def limit(self, value):

        self._limit = value

    @property
    def clean(self):

        return self._clean

    def _unique(self):

        raise NotImplementedError

    def _ensure(self):

        if not len(self) == self._limit:

            return

        key = self._unique()

        del self[key]

    def create(self, data):

        self._ensure()

        value = self._factory(data)

        return value

    def __repr__(self):

        return f'<{self.__class__.__name__}({len(self)})>'


class Array(Collect, list):

    """.."""

    def _unique(self):

        return 0

    def create(self, data):

        value = super().create(data)

        self.append(value)

        return value


class Mapping(Collect, dict):

    """.."""

    def _unique(self):

        return next(iter(self))

    def create(self, key, data):

        value = super().create(data)

        self.__setitem__(key, value)

        return value

#:
missing = type(
    'missing',
    (),
    {
        '__slots__': (),
        '__bool__': False.__bool__,
        '__iter__': ().__iter__,
        '__doc__': 'Signal for data that could be there but is not.'
    }
)()


_memories = collections.defaultdict(dict)


class Structure(cache.Entry):

    """
    Updates with incoming data and returns :class:`missing` on unwarranted
    attribute access.

    :param dict data:
        Information to store.
    :param bool direct:
        Keep the data or a copy of it.
    :param bool cache:
        Memoize this class against the first :func:`~.utils.identify` value.
        Cache-able instance creation will always return the same object, and
        just update old data with new.
    """

    __slots__ = ()

    def __new__(cls,
                /,
                data = None,
                direct = False,
                *,
                cache = False,
                **kwargs):

        memory = _memories[cls]

        cache = cache and not data is None

        if cache:

            table = utils.feel(cls)

            primary = utils.identify(data, table = table)

            self = memory.get(primary)

        else:

            self = None

        if not self:

            self = super().__new__(cls)

            if cache:

                memory[primary] = self

            root = data if direct else None

            # NOTE overwrite cls' __init__
            super().__init__(self, root, direct)

        if not (data is None or direct):

            from . import updates

            updates.structure(self, data, **kwargs)

        return self

    def __init__(self, *args, **kwargs):

        pass

    def __getattr__(self, key):

        try:

            value = super().__getattr__(key)

        except AttributeError:

            table = utils.feel(self.__class__)

            if key in table:

                return missing

            raise

        return value

    def __eq__(self, other):

        return isinstance(other, dict) and utils.equate(self, other)

    def __ne__(self, other):

        return not self.__eq__(other)

    def __repr__(self):

        return f'<{self.__class__.__name__}>'




from .cache import *

from .types import *
from .tools import *
from .access import *
from .helpers import *
from .utils import *

from . import updates


__all__ = (*cache.__all__, *types.__all__, *tools.__all__, *access.__all__,
           *helpers.__all__, *utils.__all__, 'update', 'missing')


def update(value, data, **kwargs):

    """
    Update the value with new data.
    """

    return updates.any(value, data, **kwargs)


missing = types.missing



import pathing
import collections
import copy


__all__ = ('Entry', 'Cache', 'DBCache')


class Entry:

    """
    :class:`dict`\-based attribute-accessible and read-only representation of
    data.

    .. code-block::

        >>> entry = Entry({'name': 'Pup', 'age': 6})
        >>> entry.name
        'Pup'
    """

    __slots__ = ('__data__',)

    def __init__(self, /, data = None, direct = False, **extra):

        root = (data if direct else data.copy()) if not data is None else {}

        root.update(extra)

        self.__data__ = root

    def __getitem__(self, key):

        return self.__data__[key]

    def __getattr__(self, key):

        try:

            value = self[key]

        except KeyError as error:

            raise AttributeError(*error.args) from None

        return value

    def __copy__(self):

        data = self.__data__.copy()

        fake = self.__class__(data, True)

        return fake

    def __deepcopy__(self, memo):

        fake = copy.copy(self)

        data = fake.__data__

        for (key, value) in data.items():

            data[key] = copy.deepcopy(value)

        return fake

    def __repr__(self):

        return f'<{self.__class__.__name__}>'


def update(entry, data):

    """
    Update an :class:`Entry` with the data.
    """

    entry.__data__.update(data)


class Cache(dict):

    """
    Convenient means of managing entries.

    :param int depth:
        Amount of keys required to access entries.
    :param func create:
        Transforms new data to entries.
    :param func update:
        Keeps matching entries up-to-date.

    :var entries:
        Yields all entries.
    """

    __slots__ = ('_depth', '_build',)

    _Build = collections.namedtuple('Build', 'create update')

    def __init__(self, depth, create = Entry, update = update):

        self._depth = depth

        self._build = self._Build(create, update)

    @property
    def depth(self):

        return self._depth

    def iterate(self):

        """
        Yields ``(keys, value)`` pairs.
        """

        yield from pathing.derive(self, min = self._depth, max = self._depth)

    @property
    def entries(self):

        for (keys, value) in self.iterate:

            yield value

    def _make(self, depth):

        return self.__class__(depth, *self._build)

    __final = object()

    def _seek(self, keys, final = __final):

        value = self

        abrupt = final is self.__final

        keys = tuple(keys)

        stop = len(keys) - 1

        generate = enumerate(keys)

        while True:

            yield value

            try:

                (index, key) = next(generate)

            except StopIteration:

                break

            try:

                value = value[key]

            except KeyError:

                if abrupt:

                    raise

            else:

                continue

            if index == stop:

                subvalue = final

            else:

                depth = value.depth - 1

                subvalue = self._make(depth)

            value[key] = value = subvalue

    def _wipe(self, keys):

        generate = self._seek(keys)

        values = reversed(tuple(generate))

        skip = object()

        keys = reversed((*keys, skip))

        for (value, key) in zip(values, keys):

            if key is skip:

                yield value

                continue

            yield value.pop(key)

            if len(value):

                break

    def _lead(self, *args, **kwargs):

        (*rest, value) = self._seek(*args, **kwargs)

        return value

    @classmethod
    def _flat(cls, value):

        return tuple(value.entries) if isinstance(value, cls) else (value,)

    def select(self, keys):

        """
        Get all entries against the ``keys``.
        """

        value = self._lead(keys)

        entries = self._flat(value)

        return entries

    def create(self, keys, data):

        """
        Create and hold all entries created from ``data`` against the ``keys``
        and return.
        """

        value = self._build.create(data)

        self._lead(keys, final = value)

        entries = self._flat(value)

        return entries

    def update(self, keys, data):

        """
        Update all entries with ``data`` against the ``keys`` and return.
        """

        entries = self.select(keys)

        dummies = copy.deepcopy(entries)

        for entry in entries:

            self._build.update(entry, data)

        pairs = tuple(zip(dummies, entries))

        return pairs

    def delete(self, keys):

        """
        Delete all entries with data against the ``keys`` and return.

        .. note::

            Sub-Caches will be removed as well if they appear empty.
        """

        (value, *rest) = self._wipe(keys)

        entries = self._flat(value)

        return entries

    def place(self, old, new):

        """
        Change the position of whatever is against ``old`` to ``new`` keys.
        """

        (final, *rest) = self._wipe(old)

        self._lead(new, final = final)

    def __repr__(self):

        count = len(tuple(self.entries))

        return f'<{self.__class__.__name__}[{count}]>'


class DBCache(Cache):

    """
    Treats existing funtionality as a database-like proxy.

    :param list[str] primary:
        Names of fields used to resolve primary values.
    """

    __slots__ = ('_primary',)

    def __init__(self, primary, *args, **kwargs):

        depth = len(primary)

        super().__init__(depth, *args, **kwargs)

        self._primary = primary

    @property
    def primary(self):

        return self._primary

    def _make(self, depth):

        return self.__class__(self._primary[:depth], *self._build)

    def query(self, data):

        """
        Get all values agains the ``data``\'s ``primary keys``.

        .. code-block:

        >>> cache = DBCache(('id', 'age'))
        >>> cache.query({'id': '0', 'name': 'Pup', 'age': 7})
        >>> ('0', 7)
        """

        return tuple(data[key] for key in self._primary)

    def create(self, data, /, keys = None):

        if not keys:

            keys = self.query(data)

        result = super().create(key, data)

        return result

    def update(self, data, /, keys = None):

        """
        Will also change the position of respective entries if ``keys`` don't
        match the ``primary`` ones.
        """

        if not keys:

            keys = self.query(data)

        else:

            keys = list(keys)

            fresh = keys.copy()

            for (index, name) in enumerate(self._primary):

                try:

                    key = data[name]

                except KeyError:

                    continue

                try:

                    fresh[index] = key

                except IndexError:

                    break

            if not fresh == keys:

                self.place(keys, fresh)

                keys = fresh

        return super().update(keys, data)



import enum
import itertools
import functools
import collections

from . import types
from . import utils


__all__ = ('New', 'Key', 'field')


class New:

    """
    Convenience source for new :class:`~.types.Structure` classes.

    .. code-block::

        >>> new = New()
        >>> new.Animal
        '<Animal>'
    """

    __slots__ = ('__base', '__meta', '__dict__')

    __space = {'__slots__': ()}

    def __init__(self, base = types.Structure, meta = type):

        self.__base = base

        self.__meta = meta

    def __getattr__(self, key):

        value = self.__meta(key, (self.__base,), self.__space)

        self.__dict__[key] = value

        return value


class Key(str):

    """
    Key(level, value)

    Signifies a table key.

    :param int level:
        Determines importance. For example, setting it to ``0`` would imply
        this is the primary key, ``1`` could be used for recognising, and ``2+``
        for more trivial reasons.
    :param str value:
        The name of the key.
    """

    __slots__ = ('_level',)

    def __new__(cls, /, level, value):

        self = super().__new__(cls, value)

        self._level = level

        return self

    @property
    def level(self):

        return self._level


class Table(dict):

    """
    Concentrates instructions on how to handle :class:`Structure`\s.
    """

    __slots__ = ('_groups', '_extra')

    def __init__(self, data, extra = {}):

        check = lambda key: key.level

        groups = collections.defaultdict(list)

        for (level, keys) in itertools.groupby(data, key = check):

            groups[level].extend(keys)

        self._groups = groups

        self.update(data)

        self._extra = extra

    @property
    def groups(self):

        return self._groups

    @property
    def extra(self):

        return self._extra


@functools.lru_cache(None)
def field(cls, *args, **kwargs):

    """
    Signifies a table field.

    .. code-block::

        Table(
            {
                Key(0, 'id'): field(str),
                Key(1, 'users'): field(Mapping, new.User)
            }
        )
    """

    return functools.partial(cls, *args, **kwargs)



import functools
import operator

from . import types
from . import helpers
from . import utils


__all__ = ()


def _array(store, data, equate, clean):

    proxy = store.copy()

    for data in data:

        for (index, value) in enumerate(proxy):

            if equate(value, data):

                try:

                    any(value, data)

                except TypeError:

                    pass

                del proxy[index]

                break

        else:

            value = store.create(data)

    if clean:

        for value in proxy:

            store.remove(value)


def array(value, data):

    cls = value.factory

    try:

        table = utils.feel(cls)

    except KeyError:

        equate = operator.eq

    else:

        equate = functools.partial(utils.equate, table = table)

    clean = value.clean

    return _array(value, data, equate, clean)


def _mapping(store, data, identify, clean):

    proxy = store.copy()

    for data in data:

        key = identify(data)

        try:

            value = proxy.pop(key)

        except KeyError:

            value = store.create(key, data)

            continue

        try:

            any(value, data)

        except TypeError:

            pass

    if clean:

        for key in proxy:

            del store[key]


def mapping(value, data):

    cls = value.factory

    table = utils.feel(cls)

    identify = functools.partial(utils.identify, table = table)

    clean = value.clean

    return _mapping(value, data, identify, clean)


def _structure(root, data, table, stop = False, clean = False, fill = False):

    proxy = root.copy()

    for (key, data) in data.items():

        if data is None:

            root[key] = data

            continue

        try:

            field = table[key]

        except KeyError:

            if stop:

                raise

            root[key] = data

            continue

        try:

            cls = helpers.resolve(field)

            execute = _get(cls)

        except TypeError:

            execute = None

        try:

            value = proxy.pop(key)

        except KeyError:

            if execute:

                value = field()

                execute(value, data)

            else:

                value = field(data)

        else:

            if execute:

                execute(value, data)

                continue

        root[key] = value

    if clean:

        for key in proxy:

            del root[key]

    if fill:

        accept = (types.Collect, types.Structure)

        for (key, field) in table.items():

            cls = helpers.resolve(field)

            if key in root:

                continue

            if not issubclass(cls, accept):

                continue

            root[key] = field()


def structure(value, data, **kwargs):

    table = utils.feel(value)

    return _structure(value.__data__, data, table, **kwargs)


def _get(cls):

    if issubclass(cls, types.Mapping):

        return mapping

    if issubclass(cls, types.Array):

        return array

    if issubclass(cls, types.Structure):

        return structure

    raise TypeError(cls)


def any(value, data, **kwargs):

    execute = _get(value.__class__)

    return execute(value, data, **kwargs)



import functools
import operator
import math

from . import types
from . import helpers


__all__ = ('feel', 'auto', 'locate', 'analyse', 'identify', 'equate', 'origin',
           'hosts')


def feel(value):

    """
    Get the table of the value (or class).
    """

    if not isinstance(value, type):

        value = value.__class__

    from . import tables

    return tables[value]


def auto(function):

    @functools.wraps(function)
    def wrapper(*args, table = None, **kwargs):

        if not table:

            table = feel(args[0])

        return function(table, *args, **kwargs)

    return wrapper


__foreign = object()


def locate(table, value, level, dive = 0, full = False, extra = False):

    """
    Get all keys and values against the table's level keys.
    """

    default = ()

    skip = not full

    subdive = dive - 1

    for sublevel in sorted(table.groups):

        if sublevel > level:

            break

        allow = sublevel == level

        if extra and allow:

            for apply in table.extra.get(sublevel, default):

                if isinstance(apply, str):

                    data = apply.format(**value)

                else:

                    data = apply(value)

                yield (__foreign, data)

        deny = not allow

        for key in table.groups.get(sublevel, default):

            field = table[key]

            cls = helpers.resolve(field)

            try:

                data = value[key]

            except KeyError:

                continue

            try:

                subtable = feel(cls)

            except KeyError:

                if deny and skip:

                    continue

            else:

                if dive:

                    yield from locate(
                        subtable, data, level,
                        subdive, full, extra
                    )

                continue

            yield (key, data)


@auto
def analyse(table, value, level, /, dive = 0, *, full = False, extra = False):

    """
    Get all values against the table's level keys.
    """

    if isinstance(value, types.Structure):

        value = value.__data__

    for (key, data) in locate(table, value, level, dive = dive, extra = extra):

        yield data


@auto
def identify(table, value, strict = True):

    """
    Get the first means of primary identification.
    """

    generate = analyse(value, 0, math.inf, table = table)

    try:

        value = next(generate)

    except StopIteration:

        if strict:

            raise RuntimeError('could not resolve primary key')

        if isinstance(value, types.Structure):

            value = value.__data__

    return value


@auto
def equate(table, *values):

    """
    Check whether all values are equal.
    """

    execute = functools.partial(identify, table = table, strict = False)

    return operator.eq(*map(execute, values))


@auto
def origin(table, source, target, direct = True):

    """
    Get the first attribute of source that points to the target.

    :param type source:
        Subclass for :class:`~.types.Structure`, used for searching.
    :param type target:
        Class to look for.
    :param bool direct:
        Whether to check for the target or a :class:`~.types.Collect` storing
        it.
    """

    indirect = not direct

    for (key, field) in table.items():

        check = field.func

        if issubclass(check, types.Collect):

            if direct:

                continue

            check = helpers.resolve(next(iter(field.args)))

        else:

            if indirect:

                continue

        if check is target:

            break

    else:

        raise ValueError(target)

    return key


def hosts(target, direct = True):

    """
    Get each class against its attribute that points to the value.

    :param type value:
        Class to look for in the plans.
    :param bool direct:
        Whether to check for the value or a :class:`types.Collect` storing it.
    """

    from . import tables

    for (cls, table) in tables.items():

        try:

            key = origin(cls, target, direct = direct)

        except ValueError:

            continue

        yield (cls, key)
